--
-- This module contains the logic to maintain subgroups composition when 
-- entering a battleground.
--

local Roster = AceLibrary("Roster-2.1")
local L = AceLibrary("AceLocale-2.2"):new("BiGOpt")


-- Constants

local MAX_RESTORE_ITERATIONS = 128
local MAX_SUBGROUPS = 8

-- Initialization

BiGOptSubGroups = AceLibrary("AceOO-2.0").Class()

function BiGOptSubGroups.prototype:init()
    BiGOptSubGroups.super.prototype.init(self)
    self.playerName, _ = UnitName("player")
end

-- Public interface

function BiGOptSubGroups.prototype:setConsoleRenderer(renderer)
    -- Set the console renderer
    self.renderer = renderer
end

function BiGOptSubGroups.prototype:save()
    -- Remember the subgroups composition
    self.savedRoster, self.savedLeader, self.savedRosterLength = self:getRosterSnapshot()
end

function BiGOptSubGroups.prototype:renderSaved()
    -- Render the saved roster
    if self.savedRoster ~= nil then
        self.renderer:Print("saved raid :")
        self:renderRoster(self.savedRoster)    
    else
        self.renderer:Print(L["no raid has been saved"])
    end
end

function BiGOptSubGroups.prototype:restoreLeadership()
    -- Give back the leadership to the original leader

    if not IsRaidLeader() or self.savedLeader == nil then
        return 
    end

    if self.playerName ~= self.savedLeader.name then
        local leader = Roster:GetUnitIDFromName(self.savedLeader.name)
        if leader ~= nil then
            self.renderer:Debug(L["giving back leadership to %s"], self.savedLeader.name)
            PromoteToLeader(leader)
        end
    end
end

function BiGOptSubGroups.prototype:restore()
    -- Make the current sub groups conform the saved groups

    if not IsRaidLeader() or self.savedRoster == nil then
        return
    end    

    local swapper, nbPlaced, misplaced, nbIterations
    local roster, leader, _ = self:getRosterSnapshot()
    local targetSubgroup
    local groupCounts = self:rosterGroupCounts(roster)
    local targetRoster = self:extendSavedRoster(roster)

    nbIterations = 0
    while true do
        misplaced = 0
        for playerName, player in pairs(roster) do
            if not player.subgroupOk then
                targetSubgroup = targetRoster[playerName].subgroup
                if player.subgroup == targetSubgroup then
                    player.subgroupOk = true
                else
                    if groupCounts[targetSubgroup] ~= 5 then
                        -- There is room in the target group, simply move the player
                        SetRaidSubgroup(player.index, targetSubgroup)
                        groupCounts[player.subgroup] = groupCounts[player.subgroup] - 1
                        groupCounts[targetSubgroup] = groupCounts[targetSubgroup] + 1
                        player.subgroup = targetSubgroup
                        player.subgroupOk = true
                        self.renderer:Debug(L["moving %s to group %i"], playerName, targetSubgroup)
                    else
                        -- No room left in the target group, search someone to 
                        -- swap with
                        swapper, nbPlaced = self:findSwapper(player, roster, targetRoster)
                        misplaced = misplaced + (2 - nbPlaced)
                        if swapper then
                            self:swap(player, swapper)
                        end
                    end
                end
            end
        end
        nbIterations = nbIterations + 1
        if misplaced == 0 or nbIterations == MAX_RESTORE_ITERATIONS then
            if nbIterations == MAX_RESTORE_ITERATIONS then
                self.renderer:Print(L["error: maximum number of iterations reached before having fully restored the raid group"])
            end
            break
        end
    end
end

function BiGOptSubGroups.prototype:complete()
    if self.savedRoster == nil then
        return false
    end

    local count = 0
    for unit in Roster:IterateRoster() do
        if self.savedRoster[unit.name] ~= nil then
            count = count + 1
        end
    end

    return count == self.savedRosterLength
end

function BiGOptSubGroups.prototype:clearSavedSubgroups()
    self.savedRoster = nil
    self.savedRosterLength = nil
    self.savedLeader = nil
end

-- Protected functions

function BiGOptSubGroups.prototype:getRosterSnapshot()
    local leader, unitCopy, unit
    local unitsByName = {}
    local i = 1

    while i <= 40 do
        name = GetRaidRosterInfo(i)
        if name == nil then
            break
        end
        unit = Roster:GetUnitObjectFromName(name)
        unitsByName[unit.name] = self:copyUnit(unit, i)
        if unit.rank == 2 then
            leader = self:copyUnit(unit, i)
        end
        i = i + 1
    end

    return unitsByName, leader, i - 1
end

function BiGOptSubGroups.prototype:swap(player1, player2)
    local tmp 
    SwapRaidSubgroup(player1.index, player2.index)
    self.renderer:Debug(L["swapping %s with %s"], player1.name, player2.name)
    tmp = player1.index
    player1.index = player2.index
    player2.index = tmp
    tmp = player1.subgroup
    player1.subgroup = player2.subgroup
    player2.subgroup = tmp
end

function BiGOptSubGroups.prototype:findSwapper(swappingPlayer, currentRoster, targetRoster)
    -- Find the best possible swapper for the given player, matching the 
    -- savedRoster

    local targetSubgroup = targetRoster[swappingPlayer.name].subgroup
    local acceptableSwapper

    for playerName, player in pairs(currentRoster) do
        if not player.subgroupOk and playerName ~= swappingPlayer.name then                 
            if player.subgroup == targetSubgroup then
                if targetRoster[playerName].subgroup == swappingPlayer.subgroup then
                    -- Perfect swapper, both units will be placed after swapping
                    swappingPlayer.subgroupOk = true
                    player.subgroupOk = true
                    return player, 2
                else
                    -- Acceptable swapper (only swappingPlayer will be placed
                    -- after swapping)
                    acceptableSwapper = player
                end
            end
        end
    end

    if acceptableSwapper ~= nil then
        swappingPlayer.subgroupOk = true
        return acceptableSwapper, 1
    end

    return false, 0
end

function BiGOptSubGroups.prototype:renderRoster(roster)
    for name, player in pairs(roster) do
        self.renderer:Print(string.format(L[" - %s (group %i)"], name, player.subgroup))
    end
end

function BiGOptSubGroups.prototype:copyUnit(unit, index)
    -- Return a copy of the given Roster-2.1 unit object.
    -- If the passed unit contains extended information (index and/or 
    -- subgroupOk), it is copied. If it doe not, default values are 
    -- used.

    local result = {
        name = unit.name,
        unitid = unit.unitid,
        rank = unit.rank,
        subgroup = unit.subgroup,
        online = unit.online,
    }

    if unit.index == nil then
        result.index = index
    else
        result.index = unit.index
    end

    if unit.subgroupOk == nil then
        result.subgroupOk = false
    else
        result.subgroupOk = unit.subgroupOk
    end

    return result
end

function BiGOptSubGroups.prototype:rosterGroupCounts(roster)
    local counts = {}
    for i = 1, MAX_SUBGROUPS do 
        counts[i] = 0
    end
    for name, player in pairs(roster) do
        counts[player.subgroup] = counts[player.subgroup] + 1
    end
    return counts        
end

function BiGOptSubGroups.prototype:extendSavedRoster(roster)
    -- Return the saved roster, extended with another roster.
    -- People that were not in the original raid are assigned to subgroups
    -- that have space left.

    local result = {}, playerCopy
    local counts = self:rosterGroupCounts(self.savedRoster)

    for name, player in pairs(roster) do
        if self.savedRoster[name] == nil then
            playerCopy = self:copyUnit(player)
            self:placePickup(playerCopy, counts)
            result[name] = playerCopy
        else
            result[name] = self:copyUnit(self.savedRoster[name])
        end
    end

    return result
end

function BiGOptSubGroups.prototype:placePickup(player, counts)
    -- Place a pickup player in the raid
    for i = 1, MAX_SUBGROUPS do
        if counts[i] ~= 5 then
            player.subgroup = i
            counts[i] = counts[i] + 1
            return
        end
    end

    self.renderer:Print(L["internal error, please send a bug report to the author"])
end


